//---------------------------------------------------------------------
// File: DotNetFileAsyncTransmitterBatch.cs
// 
// Summary: Implementation of an adapter framework sample adapter. 
//
// Sample: Adapter framework runtime adapter.
//
//---------------------------------------------------------------------
// This file is part of the Microsoft BizTalk Server SDK
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// This source code is intended only as a supplement to Microsoft BizTalk
// Server release and/or on-line documentation. See these other
// materials for detailed information regarding Microsoft code samples.
//
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
// KIND, WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
// PURPOSE.
//---------------------------------------------------------------------
#define SSH_ADAPTER
using System;
using System.IO;
using System.Xml;
using System.Text;
using System.Collections;
using System.Threading;
using Microsoft.BizTalk.TransportProxy.Interop;
using Microsoft.BizTalk.Component.Interop;
using Microsoft.BizTalk.Message.Interop;
using Microsoft.Samples.BizTalk.Adapter.Common;

#if SSH_ADAPTER
//
// Added code for the JNBridgePro SSH example
//      
using com.jcraft.jsch;
using com.jnbridge.examples.SSHAdapter;
using java.util;
//
// End added code
//
#endif

namespace Microsoft.BizTalk.SDKSamples.Adapters
{
	/// <summary>
	/// There is one instance of HttpTransmitterEndpoint class for each every static send port.
	/// Messages will be forwarded to this class by AsyncTransmitterBatch
	/// </summary>
	internal class DotNetFileTransmitterEndpoint : AsyncTransmitterEndpoint
	{
		private AsyncTransmitter asyncTransmitter = null;
        private string propertyNamespace;
        private static int IO_BUFFER_SIZE = 4096;

#if SSH_ADAPTER
        //
        // Added code for the JNBridgePro SSH example
        //  
        // the JSch context
        private com.jcraft.jsch.JSch jschContext;
        // Session
        private com.jcraft.jsch.Session jschSession;
        // the JSch sftp channel
        private com.jcraft.jsch.ChannelSftp sftpChannel;
        // used to ensure that JSch is initialized only once
        private bool isJSchInitialized;
        // used to store the path of the last target directory
        private string currentPath;
        //
        // End added code
        //
#endif

        public DotNetFileTransmitterEndpoint(AsyncTransmitter asyncTransmitter) : base(asyncTransmitter)
		{
			this.asyncTransmitter = asyncTransmitter;
		}

        public override void Open(EndpointParameters endpointParameters, IPropertyBag handlerPropertyBag, string propertyNamespace)
        {
            this.propertyNamespace = propertyNamespace;

#if SSH_ADAPTER
            //
            // Added code for the JNBridgePro SSH example
            //  
            JNBridgeProInit.init();
            this.currentPath = "";
            this.isJSchInitialized = false;
            //
            // End added code
            //
#endif
        }

#if SSH_ADAPTER
        //
        // Added code for the JNBridgePro SSH example
        //  
        public override IBaseMessage ProcessMessage(IBaseMessage message)
        {
            Stream source = message.BodyPart.Data;

            DotNetFileTransmitProperties props = new DotNetFileTransmitProperties(message, propertyNamespace);

            if (this.isJSchInitialized == false)
            {
                this.initJsch(props);
                this.isJSchInitialized = true;
            }

            string aPath;
            if (props.Directory != this.currentPath)
            {
                this.currentPath = props.Directory;
                // remove the drive letter from the path if it exists
                aPath = props.Directory.Substring(Path.GetPathRoot(props.Directory).Length);
                // directory actions are relative to the current working directory
                // before moving to a new directory, we'll move back to root as defined by SFTP
                // extra credit would be to map the SSH absolute path to the  Windows absolute path
                // if they're the same then there's no problem      
                this.sftpChannel.cd("/");
                try
                {
                    aPath = this.sftpChannel.realpath(aPath);
                }
                catch (Exception ex)
                {
                    throw new DotNetFileException("Unable to find path: " + aPath);
                }
                this.sftpChannel.cd(aPath);
            }

            // get the target file and trim the path
            string filePath = DotNetFileTransmitProperties.CreateFileName(message, props.Uri);
            filePath = filePath.Remove(0, filePath.LastIndexOf("\\") + 1);

            try
            {
                sbyte[] sbytes = this.readStreamToBuffer(source);
                java.io.ByteArrayInputStream inputStream = new java.io.ByteArrayInputStream(sbytes);
                this.sftpChannel.put(inputStream, filePath);
            }
            catch (Exception ex)
            {
                throw new DotNetFileException("Unable to put file: " + ex.Message, ex);
            }
            return null;
        }

        /// <summary>
        /// Create the JSch context and sftp channel
        /// </summary>
        private void initJsch(DotNetFileTransmitProperties props)
        {
            UserInfoImpl ui = new UserInfoImpl(props.PassWord, "");

            try
            {
                this.jschContext = new JSch();
        
                this.jschSession = this.jschContext.getSession(props.UserName
                    , props.HostName
                    , props.PortNumber);
                this.jschSession.setUserInfo(ui);
                this.jschSession.connect();
                com.jcraft.jsch.Channel chnl = this.jschSession.openChannel("sftp");
                chnl.connect();
                this.sftpChannel = DefaultClassWrapper.cast(chnl);
            }
            catch (Exception ex)
            {
                throw new DotNetFileException("Unable to create the JSch session: " + ex.Message, ex);
            }
        }

        public const int LARGE_DATA_THRESHHOLD = 32768;

        private  sbyte[] readStreamToBuffer(Stream stream)
        {
            byte[] buffer = new byte[LARGE_DATA_THRESHHOLD];
            int read = 0;
            int chunk;
            stream.Seek(0, SeekOrigin.Begin);
            while ((chunk = stream.Read(buffer, read, buffer.Length - read)) > 0)
            {
                read += chunk;
                //  end of buffer: check to see if more information
                if (read == buffer.Length)
                {
                    int nextByte = stream.ReadByte();
                    // EOS: done
                    if (nextByte == -1)
                    {
                        break;
                    }
                    // nope, more. resize the buffer, put in the byte just
                    // read and continue
                    byte[] newBuffer = new byte[buffer.Length * 2];
                    Array.Copy(buffer, newBuffer, buffer.Length);
                    newBuffer[read] = (byte)nextByte;
                    buffer = newBuffer;
                    read++;
                }
            }
            // move everything to a signed byte array
            // and shrink to fit
            sbyte[] ret = new sbyte[read];
            for (int ndx = 0; ndx < read; ndx++)
            {
                ret[ndx] = (sbyte)buffer[ndx]; 
            }
            return ret;
        }
        //
        // End added code
        //
#else
		/// <summary>
		/// Implementation for AsyncTransmitterEndpoint::ProcessMessage
		/// Transmit the message and optionally return the response message (for Request-Response support)
		/// </summary>
		public override IBaseMessage ProcessMessage(IBaseMessage message)
		{   		
			Stream source = message.BodyPart.Data;

		    // build url
            DotNetFileTransmitProperties props = new DotNetFileTransmitProperties(message, propertyNamespace);
            string filePath = DotNetFileTransmitProperties.CreateFileName(message, props.Uri);

		    // Create the new file
			using (FileStream fileStream = new FileStream(filePath, props.FileMode))
			{
				// Seek to the end of the file in case we're appending to existing data
				fileStream.Seek(0, SeekOrigin.End);

				source.Seek(0, SeekOrigin.Begin);

				// Copy the data from the msg to the file
				byte[] buffer = new byte[IO_BUFFER_SIZE];
				int bytesRead;
				while(0 != (bytesRead = source.Read(buffer, 0, buffer.Length)))
				{
					fileStream.Write(buffer, 0, bytesRead);
				}
			}

			return null;
		}
#endif


	}
}
